body {
    font-family: Verdana, Lato, sans-serif;
}

table {
    border-collapse: collapse;
}

tr:nth-child(2n) {
    background-color: rgba(0, 0, 0, .1);
}

th {
    border-bottom: .1em black solid;
}

th, td {
    padding: .2em .4em;
    border-right: .1em black solid;
}

th:last-of-type, td:last-of-type {
    border-right: none;
}
